@@ -65,6 +65,15 @@ class DataType(enum.Enum):
         return self.value in (DataType.INTEGER.value, DataType.FLOAT.value)
 
 
+@enum.unique
+class ArtifactStatus(enum.Enum):
+    """Artifact Status, default to USER_UPLOADED """
+
+    NO_MODEL = 'no_model'
+    SURROGATE = 'surrogate'
+    USER_UPLOADED = 'user_uploaded'
+
+
 @enum.unique
 class ModelTask(enum.Enum):
     """Supported model tasks for the Fiddler engine."""
@@ -80,6 +89,7 @@ class ModelInputType(enum.Enum):
 
     TABULAR = 'structured'
     TEXT = 'text'
+    MIXED = 'mixed'
 
 
 class AttributionExplanation(NamedTuple):
@@ -95,7 +105,14 @@ class AttributionExplanation(NamedTuple):
         """Converts a deserialized JSON format into an
         AttributionExplanation object"""
         algorithm = deserialized_json.pop('explanation_type')
-        if algorithm == 'ig' and deserialized_json['explanation'] == {}:
+        if algorithm == 'ig_flex':
+            return cls(
+                algorithm=algorithm,
+                inputs=[],
+                attributions=deserialized_json.pop('explanation'),
+                misc=deserialized_json,
+            )
+        elif algorithm == 'ig' and deserialized_json['explanation'] == {}:
             input_attr = deserialized_json.pop('explanation_ig')
             inputs, attributions = input_attr[0], input_attr[1]
         else:
@@ -157,7 +174,8 @@ class ModelDeploymentParams:
     """Holds configuration information for a model packaged as a container."""
 
     def __init__(
-        self, image: str,
+        self,
+        image: str,
     ):
         self.image = image
 
@@ -204,14 +222,17 @@ class Column:
         self.value_range_min = value_range_min
         self.value_range_max = value_range_max
 
-        if (
-            self.data_type.value != DataType.CATEGORY.value
-            and self.possible_values is not None
-        ):
-            raise ValueError(
-                f'Do not pass `possible_values` for '
-                f'non-categorical {self.data_type} data type.'
-            )
+        # this check is not needed. User can switch from
+        # category to str, then from str to category in the UI.
+        # so we should not be deleting possible-values
+        # if (
+        #     self.data_type.value != DataType.CATEGORY.value
+        #     and self.possible_values is not None
+        # ):
+        #     raise ValueError(
+        #         f'Do not pass `possible_values` for '
+        #         f'non-categorical {self.data_type} data type.'
+        #     )
         inappropriate_value_range = not self.data_type.is_numeric() and not (
             self.value_range_min is None and self.value_range_max is None
         )
@@ -650,8 +671,8 @@ class DatasetInfo:
 
             # resolve possible_values
             if col_type.value == DataType.CATEGORY.value:
-                assert a_column.possible_values is not None
-                assert b_column.possible_values is not None
+                assert a_column.possible_values is not None  # nosec
+                assert b_column.possible_values is not None  # nosec
                 possible_values = list(
                     set(a_column.possible_values) | set(b_column.possible_values)
                 )
@@ -729,7 +750,9 @@ class DatasetInfo:
         return self.columns[self._col_id_from_name(item)]
 
     def __setitem__(self, key, value):
-        assert isinstance(value, Column), 'Must set column to be a ' '`Column` object'
+        assert isinstance(value, Column), (  # nosec
+            'Must set column to be a ' '`Column` object'
+        )
         self.columns[self._col_id_from_name(key)] = value
 
     def __delitem__(self, key):
@@ -755,6 +778,7 @@ class ModelInfo:
     :param description: A user-facing description of the model.
     :param mlflow_params: MLFlow parameters.
     :param model_deployment_params: Model Deployment parameters.
+    :param artifact_status: Status of the model artifact
     :param **kwargs: Additional information about the model to store as `misc`.
     """
 
@@ -773,6 +797,7 @@ class ModelInfo:
         datasets: Optional[List[str]] = None,
         mlflow_params: Optional[MLFlowParams] = None,
         model_deployment_params: Optional[ModelDeploymentParams] = None,
+        artifact_status: Optional[ArtifactStatus] = None,
         **kwargs,
     ):
         self.display_name = display_name
@@ -788,6 +813,7 @@ class ModelInfo:
         self.datasets = datasets
         self.mlflow_params = mlflow_params
         self.model_deployment_params = model_deployment_params
+        self.artifact_status = artifact_status
         self.misc = kwargs
 
     def to_dict(self):
@@ -816,6 +842,8 @@ class ModelInfo:
             res['mlflow'] = self.mlflow_params.to_dict()
         if self.model_deployment_params is not None:
             res['model_deployment'] = self.model_deployment_params.to_dict()
+        if self.artifact_status is not None:
+            res['artifact_status'] = self.artifact_status.value
         return {**res, **self.misc}
 
     def get_input_names(self):
@@ -861,7 +889,7 @@ class ModelInfo:
         self,
     ) -> Dict[str, Union[str, pandas.api.types.CategoricalDtype]]:
         """Get a dictionary describing the pandas datatype of every decision
-        column. """
+        column."""
         return _get_field_pandas_dtypes(self.decisions)
 
     def get_target_pandas_dtypes(
@@ -882,6 +910,10 @@ class ModelInfo:
         model_task = ModelTask(deserialized_json.pop('model-task'))
         inputs = [Column.from_dict(c) for c in deserialized_json.pop('inputs')]
         outputs = [Column.from_dict(c) for c in deserialized_json.pop('outputs')]
+        if 'artifact_status' in deserialized_json:
+            artifact_status = ArtifactStatus(deserialized_json.pop('artifact_status'))
+        else:
+            artifact_status = None
         if 'metadata' in deserialized_json:
             metadata = [Column.from_dict(c) for c in deserialized_json.pop('metadata')]
         else:
@@ -930,14 +962,15 @@ class ModelInfo:
             datasets=datasets,
             mlflow_params=mlflow_params,
             model_deployment_params=model_deployment_params,
+            artifact_status=artifact_status,
             **deserialized_json,
         )
 
-    @classmethod
+    @classmethod  # noqa: C901
     def from_dataset_info(
         cls,
         dataset_info: DatasetInfo,
-        target: str,
+        target: Optional[str],
         features: Optional[Sequence[str]] = None,
         metadata_cols: Optional[Sequence[str]] = None,
         decision_cols: Optional[Sequence[str]] = None,
@@ -945,7 +978,7 @@ class ModelInfo:
         description: Optional[str] = None,
         input_type: ModelInputType = ModelInputType.TABULAR,
         model_task: Optional[ModelTask] = None,
-        mlflow_params: Optional[MLFlowParams] = None,
+        outputs: Optional[Sequence[str]] = None,
         model_deployment_params: Optional[ModelDeploymentParams] = None,
     ):
         """Produces a ModelInfo for a model trained on a dataset.
@@ -967,6 +1000,7 @@ class ModelInfo:
             model. If not explicitly provided, this will be inferred from the
             data type of the target variable.
         :param mlflow_params: MLFlow parameters.
+        :param outputs model output parameters
         :param model_deployment_params: Model Deployment parameters.
 
         :returns A ModelInfo object.
@@ -1020,65 +1054,78 @@ class ModelInfo:
             if decision_cols and col_name in decision_cols:
                 decisions.append(column.copy())
 
-        # determine target column
-        try:
-            target_column = dataset_info[target]
-        except KeyError:
-            raise ValueError(f'Target "{target}" not found in dataset.')
+        target_column = None
+        if target:
+            # determine target column
+            try:
+                target_column = dataset_info[target]
+            except KeyError:
+                raise ValueError(f'Target "{target}" not found in dataset.')
 
-        # infer task type, outputs, and target levels from target
-        if target_column.data_type.value == DataType.BOOLEAN.value:
-            target_levels = [False, True]
-            outputs = [
-                Column(
-                    name=f'probability_{target_column.name}_True',
-                    data_type=DataType.FLOAT,
-                    is_nullable=False,
-                    value_range_min=0.0,
-                    value_range_max=1.0,
+        if outputs:
+            output_columns = []
+            for output in outputs:
+                output_columns.append(
+                    Column(
+                        name=output,
+                        data_type=DataType.FLOAT,
+                        is_nullable=False,
+                    )
                 )
-            ]
-            if model_task is None:
-                model_task = ModelTask.BINARY_CLASSIFICATION
-        elif target_column.data_type.value == DataType.CATEGORY.value:
-            target_levels = target_column.possible_values
-            if model_task is None:
-                if len(target_levels) == 2:
-                    model_task = ModelTask.BINARY_CLASSIFICATION
-                else:
-                    model_task = ModelTask.MULTICLASS_CLASSIFICATION
-            if model_task.value == ModelTask.BINARY_CLASSIFICATION.value:
-                outputs = [
-                    Column(
-                        name=f'probability_{target_levels[1]}',
-                        data_type=DataType.FLOAT,
-                        is_nullable=False,
-                        value_range_min=0.0,
-                        value_range_max=1.0,
-                    )
-                ]
-            else:
-                outputs = [
-                    Column(
-                        name=f'probability_{level}',
-                        data_type=DataType.FLOAT,
-                        is_nullable=False,
-                        value_range_min=0.0,
-                        value_range_max=1.0,
-                    )
-                    for level in target_levels
-                ]
         else:
-            target_levels = None
-            outputs = [
-                Column(
-                    name=f'predicted_{target_column.name}',
-                    data_type=DataType.FLOAT,
-                    is_nullable=False,
-                )
-            ]
-            if model_task is None:
-                model_task = ModelTask.REGRESSION
+            # infer task type, outputs, and target levels from target
+            if target_column.data_type.value == DataType.BOOLEAN.value:
+                target_levels = [False, True]
+                output_columns = [
+                    Column(
+                        name=f'probability_{target_column.name}_True',
+                        data_type=DataType.FLOAT,
+                        is_nullable=False,
+                        value_range_min=0.0,
+                        value_range_max=1.0,
+                    )
+                ]
+                if model_task is None:
+                    model_task = ModelTask.BINARY_CLASSIFICATION
+            elif target_column.data_type.value == DataType.CATEGORY.value:
+                target_levels = target_column.possible_values
+                if model_task is None:
+                    if len(target_levels) == 2:
+                        model_task = ModelTask.BINARY_CLASSIFICATION
+                    else:
+                        model_task = ModelTask.MULTICLASS_CLASSIFICATION
+                if model_task.value == ModelTask.BINARY_CLASSIFICATION.value:
+                    output_columns = [
+                        Column(
+                            name=f'probability_{target_levels[1]}',
+                            data_type=DataType.FLOAT,
+                            is_nullable=False,
+                            value_range_min=0.0,
+                            value_range_max=1.0,
+                        )
+                    ]
+                else:
+                    output_columns = [
+                        Column(
+                            name=f'probability_{level}',
+                            data_type=DataType.FLOAT,
+                            is_nullable=False,
+                            value_range_min=0.0,
+                            value_range_max=1.0,
+                        )
+                        for level in target_levels
+                    ]
+            else:
+                target_levels = None
+                output_columns = [
+                    Column(
+                        name=f'predicted_{target_column.name}',
+                        data_type=DataType.FLOAT,
+                        is_nullable=False,
+                    )
+                ]
+                if model_task is None:
+                    model_task = ModelTask.REGRESSION
 
         datasets = None
         if dataset_info.dataset_id is not None:
@@ -1090,12 +1137,12 @@ class ModelInfo:
             input_type=input_type,
             model_task=model_task,
             inputs=inputs,
-            outputs=outputs,
+            outputs=output_columns,
             metadata=metadata,
             decisions=decisions,
             datasets=datasets,
-            targets=[target_column],
-            mlflow_params=mlflow_params,
+            targets=[target_column] if target_column else [],
+            mlflow_params=None,
             model_deployment_params=model_deployment_params,
         )
 
