@@ -9,24 +9,25 @@ import pickle
 import shutil
 import tempfile
 import textwrap
+import time
 from collections import namedtuple
 from pathlib import Path
 from typing import Any, Dict, Iterable, List, Optional, Sequence, Union
-from urllib.parse import urlparse
 
-import numpy as np
 import pandas as pd
 import requests
 import yaml
 
 from .core_objects import (
+    ArtifactStatus,
     AttributionExplanation,
+    Column,
     DatasetInfo,
+    DataType,
     ModelInfo,
-    ModelInputType,
-    ModelTask,
     MulticlassAttributionExplanation,
 )
+from .model_info_validator import ModelInfoValidator
 from .utils import _df_to_dict, _try_series_retype, df_from_json_rows
 
 LOG = logging.getLogger()
@@ -109,7 +110,6 @@ class FiddlerApi:
         self.auth_header = {AUTH_HEADER_KEY: f'Bearer {auth_token}'}
         self.streaming_header = {STREAMING_HEADER_KEY: 'application/jsonlines'}
         self.verbose = verbose
-        LOG.info(f'url ={url}')
         try:
             _ = self.list_projects()
         except requests.exceptions.ConnectionError:
@@ -133,6 +133,7 @@ class FiddlerApi:
             'executor',
             'explain',
             'explain_by_row_id',
+            'fairness',
             'feature_importance',
             'generate',
             'new_project',
@@ -145,6 +146,25 @@ class FiddlerApi:
         else:
             return {ROUTING_HEADER_KEY: 'data_service'}
 
+    def _call_executor_service(
+        self,
+        path: List[str],
+        json_payload: Any = None,
+        files: Optional[List[Path]] = None,
+        is_get_request: bool = False,
+        stream: bool = False,
+    ):
+        no_auth_onebox = False
+        try:
+            if self.url == 'http://localhost:6100':
+                no_auth_onebox = True
+                self.url = 'http://localhost:5100'
+
+            return self._call(path, json_payload, files, is_get_request, stream)
+        finally:
+            if no_auth_onebox:
+                self.url = 'http://localhost:6100'
+
     def _call(
         self,
         path: List[str],
@@ -528,10 +548,13 @@ class FiddlerApi:
         else:
             dataset_info = copy.deepcopy(dataset_info)
 
-        def get_df_from_split(split):
+        def get_df_from_split(split, fiddler_id=include_fiddler_id):
+            column_names = dataset_info.get_column_names()
+            if fiddler_id:
+                column_names.insert(0, '__fiddler_id')
             dataset_rows = self._query_dataset(
                 dataset_id,
-                fields=dataset_info.get_column_names(),
+                fields=column_names,
                 max_rows=max_rows,
                 split=split,
                 sampling=sampling,
@@ -640,19 +663,19 @@ class FiddlerApi:
 
         :returns: Server response for deletion action.
         """
+        payload = {
+            'project_id': project_id,
+            'model_id': model_id,
+            'delete_prod': delete_prod,
+            'delete_pred': delete_pred,
+        }
+
+        path = ['delete_model', self.org_id, project_id, model_id]
         try:
-            path = ['v1', 'model', 'delete', self.org_id]
-            payload = {
-                'project_id': project_id,
-                'model_id': model_id,
-                'delete_prod': delete_prod,
-                'delete_pred': delete_pred,
-            }
             result = self._call(path, json_payload=payload)
         except Exception:
-            path = ['delete_model', self.org_id, project_id, model_id]
-            result = self._call(path)
-
+            # retry on error
+            result = self._call(path, json_payload=payload)
         return result
 
     def delete_model_artifacts(self, project_id: str, model_id: str):
@@ -695,6 +718,7 @@ class FiddlerApi:
         csv_file_paths: List[Path],
         split_test: Optional[bool] = None,
         dataset_info: Optional[DatasetInfo] = None,
+        do_import: Optional[bool] = None,
     ):
         """Uploads a CSV file to the Fiddler platform."""
         payload = dict(dataset_name=dataset_id)
@@ -702,6 +726,8 @@ class FiddlerApi:
             payload['split_test'] = split_test
         if dataset_info is not None:
             payload['dataset_info'] = dict(dataset=dataset_info.to_dict())
+        if do_import is not None:
+            payload['do_import'] = do_import
 
         try:
             path = ['v1', 'dataset', 'upload', self.org_id]
@@ -722,6 +748,23 @@ class FiddlerApi:
 
         return res
 
+    def _import_model_predictions(
+        self,
+        project_id: str,
+        dataset_id: str,
+        model_id: str,
+        columns: Sequence[Dict],
+        csv_file_paths: List[Path],
+    ):
+        """Uploads model predictions to Fiddler platform."""
+        payload = dict(dataset=dataset_id)
+        payload['model'] = model_id
+        payload['columns'] = columns
+
+        path = ['import_model_predictions', self.org_id, project_id]
+        result = self._call(path, json_payload=payload, files=csv_file_paths)
+        return result
+
     def upload_dataset(
         self,
         dataset: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
@@ -822,15 +865,15 @@ class FiddlerApi:
             new_schema.files = [path.name for path in csv_paths]
 
             # upload the CSV
-            LOG.info(f'[{dataset_id}] dataset upload: uploading csv')
-            self._upload_dataset_csvs(
-                dataset_id, csv_paths, split_test=split_test, dataset_info=new_schema
+            LOG.info(f'[{dataset_id}] dataset upload: upload and import csv')
+            res = self._upload_dataset_csvs(
+                dataset_id,
+                csv_paths,
+                split_test=split_test,
+                dataset_info=new_schema,
+                do_import=True,
             )
-
-        # trigger dataset upload
-        LOG.info(f'[{dataset_id}] dataset upload: ' f'triggering import of dataset')
-        res = self._db_import(dataset_id)
-        return res
+            return res
 
     def upload_dataset_from_dir(self, dataset_id: str, dataset_dir: Path):
         logging.info(f'uploading dataset from dir: {dataset_dir}')
@@ -844,8 +887,9 @@ class FiddlerApi:
             files = dataset_dir.glob('*.csv')
             csv_files = [x for x in files if x.is_file()]
             logging.info(f'Found CSV file {csv_files}')
-            self._upload_dataset_csvs(dataset_id, csv_files, False, dataset_info)
-            result = self._db_import(dataset_id)
+            result = self._upload_dataset_csvs(
+                dataset_id, csv_files, False, dataset_info, True
+            )
             LOG.info(f'Dataset uploaded {result}')
 
     def run_model(
@@ -873,13 +917,13 @@ class FiddlerApi:
 
         try:
             path = ['v1/execution', 'run', self.org_id]
-            res = self._call(path, json_payload=payload)
+            res = self._call_executor_service(path, json_payload=payload)
         except Exception:
             payload.pop('project_id')
             payload.pop('model_id')
 
             path = ['execute', self.org_id, project_id, model_id]
-            res = self._call(path, json_payload=payload)
+            res = self._call_executor_service(path, json_payload=payload)
         return pd.DataFrame(res)
 
     def run_explanation(
@@ -930,14 +974,14 @@ class FiddlerApi:
 
         try:
             path = ['v1/point_explanation', 'run', self.org_id]
-            res = self._call(path, json_payload=payload)
+            res = self._call_executor_service(path, json_payload=payload)
 
         except Exception:
             payload.pop('project_id')
             payload.pop('model_id')
 
             path = ['explain', self.org_id, project_id, model_id]
-            res = self._call(path, json_payload=payload)
+            res = self._call_executor_service(path, json_payload=payload)
 
         explanations_list = res['explanations']
         # TODO: enable more robust check for multiclass explanations
@@ -975,8 +1019,9 @@ class FiddlerApi:
         :param dataset_splits: If specified, importance will only be computed
             over these splits. Otherwise, all splits will be used. Only a
             single split is currently supported.
+        :param slice_query: A special SQL query.
         :param kwargs: Additional parameters to be passed to the importance
-            algorithm. For example, `n_iterations`, `n_references`, and
+            algorithm. For example, `n_inputs`, `n_iterations`, `n_references`,
             `ci_confidence_level`.
         :return: A named tuple with the explanation results.
         """
@@ -1013,6 +1058,79 @@ class FiddlerApi:
         res = namedtuple('FeatureImportanceResults', res)(**res)
         return res
 
+    def run_fairness(
+        self,
+        project_id: str,
+        model_id: str,
+        dataset_id: str,
+        protected_features: list,
+        positive_outcome: Union[str, int],
+        slice_query: Optional[str] = None,
+        score_threshold: Optional[float] = 0.5,
+    ) -> Dict[str, Any]:
+        """Get fairness metrics for a model over a dataset.
+
+        :param project_id: The unique identifier of the model's project on the
+            Fiddler engine.
+        :param model_id: The unique identifier of the model in the specified
+            project on the Fiddler engine.
+        :param dataset_id: The unique identifier of the dataset in the
+            Fiddler engine.
+        :param protected_features: List of protected features
+        :param positive_outcome: Name or value of the positive outcome
+        :param slice_query: If specified, slice the data.
+        :param score_threshold: positive score threshold applied to get outcomes
+        :return: A dictionary with the fairness metrics, technical_metrics,
+         model_outcomes and correlation_features
+        """
+        if isinstance(protected_features, str):
+            protected_features = [protected_features]
+
+        payload = dict(
+            subject='fairness',
+            project_id=project_id,
+            model_id=model_id,
+            dataset_id=dataset_id,
+            protected_features=protected_features,
+            slice_query=slice_query,
+            score_threshold=score_threshold,
+            positive_outcome=positive_outcome,
+        )
+
+        try:
+            path = ['v1/model_fairness', 'run', self.org_id]
+            res = self._call(path, json_payload=payload)
+        except Exception:
+            payload.pop('subject')
+            payload.pop('project_id')
+            payload.pop('model_id')
+
+            path = ['fairness', self.org_id, project_id, model_id]
+            res = self._call(path, json_payload=payload)
+        return res
+
+    def get_correlation_matrix(self, dataset_id, features):
+        """
+
+        :param dataset_id: The unique identifier of the dataset in the
+            Fiddler engine.
+        :param features: list of features to compute mutual information.
+        :return: a dictionary of mutual information w.r.t the given features.
+        """
+        if isinstance(features, str):
+            features = [features]
+        if not isinstance(features, list):
+            raise ValueError(
+                f'Invalid type: {type(features)}. Argument features has to be a list'
+            )
+        correlation = {}
+        for col_name in features:
+            payload = dict(org=self.org_id, dataset=dataset_id, col_name=col_name)
+            path = ['dataset_mutual_information', self.org_id, dataset_id]
+            res = self._call(path, json_payload=payload)
+            correlation[col_name] = res
+        return correlation
+
     def create_project(self, project_id: str):
         """Create a new project.
 
@@ -1045,6 +1163,7 @@ class FiddlerApi:
         features: Optional[List[str]] = None,
         train_splits: Optional[List[str]] = None,
         model_id: str = 'fiddler_generated_model',
+        model_info: Optional[ModelInfo] = None,
     ):
         """Trigger auto-modeling on a dataset already uploaded to Fiddler.
 
@@ -1099,24 +1218,18 @@ class FiddlerApi:
             features.remove(target)
 
         payload = {
-            'project_id': project_id,
-            'dataset_id': dataset_id,
-            'model_id': model_id,
+            'dataset': dataset_id,
+            'model_name': model_id,
             'source': source,
             'inputs': features,
             'targets': [target],
         }
 
-        try:
-            path = ['v1', 'training', 'run', self.org_id]
-            result = self._call(path, json_payload=payload)
-        except Exception:
-            payload.pop('project_id')
-            payload['dataset'] = payload.pop('dataset_id')
-            payload['model_name'] = payload.pop('model_id')
+        if model_info:
+            payload['model_info'] = dict(model=model_info.to_dict())
 
-            path = ['generate', self.org_id, project_id]
-            result = self._call(path, json_payload=payload)
+        path = ['generate', self.org_id, project_id]
+        result = self._call_executor_service(path, json_payload=payload)
         return result
 
     def upload_model_sklearn(
@@ -1295,7 +1408,6 @@ class FiddlerApi:
             )
 
             try:
-                raise ValueError('foo')
                 endpoint_path = ['v1', 'model', 'upload', self.org_id]
                 result = self._call(
                     endpoint_path, json_payload=payload, files=[Path(tmp) / 'files.tar']
@@ -1348,7 +1460,9 @@ class FiddlerApi:
             )
 
     def await_model_deployment(
-        self, project_id: str, model_id: str,
+        self,
+        project_id: str,
+        model_id: str,
     ):
         """
         Await model deployment.
@@ -1356,11 +1470,16 @@ class FiddlerApi:
         assert ' ' not in model_id, 'The model identifier should not contain whitespace'
 
         endpoint_path = ['model_await_deployment', self.org_id]
-        payload = dict(project=project_id, model=model_id,)
+        payload = dict(
+            project=project_id,
+            model=model_id,
+        )
         return self._call(endpoint_path, json_payload=payload)
 
     def delete_model_deployment(
-        self, project_id: str, model_id: str,
+        self,
+        project_id: str,
+        model_id: str,
     ):
         """
         Delete the model deployment.
@@ -1368,20 +1487,40 @@ class FiddlerApi:
         assert ' ' not in model_id, 'The model identifier should not contain whitespace'
 
         endpoint_path = ['model_delete_deployment', self.org_id]
-        payload = dict(project=project_id, model=model_id,)
+        payload = dict(
+            project=project_id,
+            model=model_id,
+        )
         return self._call(endpoint_path, json_payload=payload)
 
-    def setup_monitoring(
+    def create_surrogate_model(
         self, project_id, name, baseline_df, dataset_info, target, features
     ):
+        """
+        Uploads a baseline to fiddler, and creates a surrogate model
+        and sets up monitoring on this stream
+
+        :param project_id: id of the project
+        :param name: name to be used for the dataset and model
+        :param baseline_df: baseline Dataframe
+        :param dataset_info: schema for the baseline
+        :param target: target for the generated model
+        :param features: input features to the model
+        :return: location of the newly created model
+        """
+        print('Validating inputs ...')
         try:
             self.get_dataset_info(name)
             raise ValueError(f'name already in used: {name}')
         except Exception:
             LOG.info('dataset not found, creating it')
 
+        print('Uploading dataset ...')
+
+        dataset_info.dataset_id = name
         self.upload_dataset(baseline_df, name, split_test=False, info=dataset_info)
         train_splits = ['data']
+        print('Generating surrogate model ...')
         self.create_model(
             project_id,
             name,
@@ -1390,6 +1529,91 @@ class FiddlerApi:
             train_splits=train_splits,
             model_id=name,
         )
+        print('Triggering model predictions ...')
+        self.trigger_model_predictions(project_id, name, name)
+
+        project_url = self.url.replace('host.docker.internal', 'localhost', 1)
+        return (
+            f'Surrogate model successfully setup on Fiddler. \n '
+            f'Visit {project_url}/projects/{project_id} '
+        )
+
+    def create_model_from_prediction_log(
+        self,
+        project_id,
+        model_id,
+        prediction_log_df,
+        features,
+        outputs,
+        model_task,
+        dataset_info=None,
+    ):
+        """
+        Creates a model based on the prediction log specified.
+
+        :param project_id: id of the project
+        :param model_id: name to be used for the dataset and model
+        :param prediction_log_df: production log Dataframe
+        :param features: input features to the model
+        :param outputs: model outputs
+        :param model_task:
+        :param dataset_info: schema for the baseline
+        :param model_task: type of the model eg: binary_classification
+        :return: location of the newly created model
+        """
+        print('Validating inputs ...')
+        try:
+            self.get_dataset_info(model_id)
+            raise ValueError(f'name already in used: {model_id}')
+        except Exception:
+            LOG.info('dataset not found, creating it')
+
+        print('Uploading dataset ...')
+        baseline_df = prediction_log_df.drop(columns=outputs)
+
+        if not dataset_info:
+            dataset_info = DatasetInfo.from_dataframe(
+                baseline_df, max_inferred_cardinality=1000
+            )
+
+        dataset_info.dataset_id = model_id
+        self.upload_dataset(baseline_df, model_id, split_test=False, info=dataset_info)
+        print('Setting up monitoring without model artifact ...')
+        package_py = (
+            'def get_model():\n\traise RuntimeError("Model artifact not available")'
+        )
+        with tempfile.TemporaryDirectory() as tmp:
+            tmp_dir = Path(tmp)
+            model_info = ModelInfo.from_dataset_info(
+                dataset_info=dataset_info,
+                target=None,
+                features=features,
+                display_name=model_id,
+                description='Model-less monitoring',
+                model_task=model_task,
+                outputs=outputs,
+            )
+            model_info.framework = None
+            model_info.artifact_status = ArtifactStatus.NO_MODEL
+            package_py_file = tmp_dir / 'package.py'
+            with open(package_py_file, mode='w') as fid:
+                fid.write(package_py)
+            self.upload_model_custom(tmp_dir, model_info, project_id, model_id)
+
+        output_columns = [Column('__dataset_fiddler_id', DataType.INTEGER).to_dict()]
+        for output in outputs:
+            output_columns.append(Column(output, DataType.FLOAT).to_dict())
+
+        with tempfile.TemporaryDirectory() as tmp:
+            tmp_dir = Path(tmp)
+            output_df = prediction_log_df[outputs]
+            output_df.insert(0, '__dataset_fiddler_id', range(1, 1 + len(output_df)))
+            path = tmp_dir / 'data.csv'
+            output_df.to_csv(path, index=False, header=None)
+            self._import_model_predictions(
+                project_id, model_id, model_id, output_columns, [path]
+            )
+
         project_url = self.url.replace('host.docker.internal', 'localhost', 1)
         return (
             f'Monitoring successfully setup on Fiddler. \n '
@@ -1417,18 +1641,10 @@ class FiddlerApi:
     ):
         """Makes the Fiddler service compute and cache model predictions on a
         dataset."""
-        payload = dict(project_id=project_id, model_id=model_id, dataset_id=dataset_id)
-
-        try:
-            path = ['v1', 'execution', 'run', self.org_id]
-            result = self._call(path, json_payload=payload)
-        except Exception:
-            payload.pop('project_id')
-            payload['dataset'] = payload.pop('dataset_id')
-
-            result = self._call(
-                ['dataset_predictions', self.org_id, project_id], payload
-            )
+        payload = dict(model=model_id, dataset=dataset_id)
+        result = self._call_executor_service(
+            ['dataset_predictions', self.org_id, project_id], payload
+        )
 
         return result
 
@@ -1471,3 +1687,167 @@ class FiddlerApi:
         result = self._call(path, event)
 
         return result
+
+    def add_monitoring_config(
+        self,
+        config_info: dict,
+        project_id: Optional[str] = None,
+        model_id: Optional[str] = None,
+    ):
+        """Adds a config for either an entire org, or project or a model.
+        Here's a sample config:
+        {
+            'min_bin_value': 3600, # possible values 300, 3600, 7200, 43200, 86400, 604800 secs
+            'time_ranges': ['Day', 'Week', 'Month', 'Quarter', 'Year'],
+            'default_time_range': 7200,
+            'tag': 'anything you want'
+        }
+        """
+        path = ['monitoring_setup', self.org_id]
+        if project_id:
+            path.append(project_id)
+        if model_id:
+            if not project_id:
+                raise ValueError(
+                    'We need to have a project_id when a model is specified'
+                )
+            path.append(model_id)
+
+        result = self._call(path, config_info)
+        return result
+
+    def publish_events_log(
+        self,
+        project_id: str,
+        model_id: str,
+        logs: pd.DataFrame,
+        force_publish: Optional[bool] = None,
+        ts_column: Optional[str] = None,
+        default_ts: Optional[int] = None,
+    ):
+        """
+        Publishes prediction log to Fiddler Service.
+        :param project_id:    the project to which the model whose events are
+                              being published belongs
+        :param model_id:      the model whose events are being published
+        :param logs:          data frame of event details, such as features
+                              and predictions.
+        :param force_publish: continue with publish even if all input and
+                              output columns not in log.
+        :param ts_column:     column to extract time stamp value from.
+        :param default_ts:    default time stamp to use if ts_column not specified.
+        """
+        try:
+            model_info = self.get_model_info(project_id, model_id)
+        except RuntimeError:
+            raise RuntimeError(
+                f'Did not find ModelInfo for project "{project_id}" and model "{model_id}".'
+            )
+
+        log_columns = [c for c in list(logs.columns)]
+        in_columns = [c.name for c in model_info.inputs]
+        in_not_found = [c for c in in_columns if c not in log_columns]
+        out_columns = [c.name for c in model_info.outputs]
+        out_not_found = [c for c in out_columns if c not in log_columns]
+        if (out_not_found or in_not_found) and not force_publish:
+            raise ValueError(
+                f'Model output columns "{out_not_found}" or input columns "{in_not_found}"'
+                f'not found in logs. If this is expected try again with force_publish=True.'
+            )
+
+        payload = dict()
+        if ts_column is not None:
+            if ts_column not in log_columns:
+                raise ValueError(f'Did not find {ts_column} in the logs columns.')
+            payload['ts_column'] = ts_column
+        else:
+            if default_ts is None:
+                default_ts = int(round(time.time() * 1000))
+            payload['default_ts'] = default_ts
+
+        include_index = logs.index.name is not None
+        with tempfile.TemporaryDirectory() as tmp:
+            csv_paths = list()
+            csv_path = Path(tmp) / 'log.csv'
+            logs.to_csv(csv_path, index=include_index)
+            csv_paths.append(csv_path)
+            path = ['publish_events_log', self.org_id, project_id, model_id]
+            result = self._call(path, json_payload=payload, files=csv_paths)
+            return result
+
+    def publish_parquet_s3(
+        self,
+        project_id: str,
+        model_id: str,
+        parquet_file: str,
+        auth_context: Optional[Dict[str, Any]] = None,
+        ts_column: Optional[str] = None,
+        default_ts: Optional[int] = None,
+    ):
+        """
+        Publishes parquet events file from S3 to Fiddler instance. Experimental and may be expanded in the future.
+
+        :param project_id:    the project to which the model whose events are
+                              being published belongs
+        :param model_id:      the model whose events are being published
+        :param parquet_file:  s3_uri for parquet file to be published
+        :param auth_context:  Dictionary containing authorization for AWS. List of expected keys are
+                              ['aws_access_key_id', 'aws_secret_access_key', 'aws_session_token']
+                              with 'aws_session_token' being applicable to the AWS account being used.
+        :param ts_column:     column to extract time stamp value from.
+        :param default_ts:    default time stamp to use if ts_column not specified.
+        """
+        payload = dict()
+        payload['file_path'] = parquet_file
+        payload['auth_context'] = auth_context
+        if ts_column is not None:
+            payload['ts_column'] = ts_column
+        else:
+            if default_ts is None:
+                default_ts = int(round(time.time() * 1000))
+            payload['default_ts'] = default_ts
+
+        publish_path = ['publish_parquet_file', self.org_id, project_id, model_id]
+        publish_result = self._call(publish_path, json_payload=payload, stream=True)
+        for res in publish_result:
+            print(res)
+
+    def register_model(self, project_id, model_id, dataset_id, model_info):
+        """
+        Register a model in fiddler. This will generate a surrogate model,
+        which can be replaced later with original model.
+
+        Note: This method can take a while if the dataset is large. It is
+        recommended to call register_model on a smaller representative
+        dataset, before trying out on larger dataset.
+
+        :param project_id: id of the project
+        :param model_id: name to be used for the dataset and model
+        :param dataset_id: id of the dataset to be used
+        :param model_info: model info
+        """
+        print('Loading dataset info ...')
+        dataset_info = self.get_dataset_info(dataset_id)
+
+        print('Validating model info ...')
+        ModelInfoValidator(model_info, dataset_info).validate()
+
+        print('Generating model ...')
+        self.create_model(
+            project_id,
+            dataset_id,
+            target=model_info.targets[0].name,
+            features=[],
+            train_splits=None,
+            model_id=model_id,
+            model_info=model_info,
+        )
+
+        print('Analysing model predictions ...')
+        self.trigger_model_predictions(project_id, model_id, dataset_id)
+
+        project_url = self.url.replace('host.docker.internal', 'localhost', 1)
+        return (
+            f'Model successfully registered on Fiddler. \n '
+            f'Visit {project_url}/projects/{project_id} '
+        )
